/* Name: hidtool.c
 * Project Radioactive@Home
 * Author: Krzysztof 'krzyszp' Piszczek based on:
 * Copyright: (c) 2011 by Radioactive@Home team, Fundacja BOINC Polska
 * Creation Date: 2011-06-18
 * License: GNU GPL v2
 * based on functions from hid-data example by Christian Starkjohann, OBJECTIVE DEVELOPMENT Software GmbH
 * This Revision: $Id: hidtool.c 101 2011-06-17 19:04:32Z cs $
 *
 */

#define LAST_CHANGED_REVISION "$Rev: 260 $"


#ifdef _WIN32
#include "boinc_win.h"
#include "windows.h" // for windows only
#include <conio.h> // for windows only
#else
#include "config.h"
#include <cstdio>
#include <cctype>
#include <ctime>
#include <cstring>
#include <cstdlib>
#include <csignal>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#endif
#include "str_util.h"
#include "util.h"
#include "filesys.h"
#include "boinc_api.h"
#include "mfile.h"
// do HIDu
#include "hiddata.h"
#include "./firmware/usbconfig.h"  /* for device VID, PID, vendor name and product name */
#include "./xmlParser/xmlParser.h" /*Frank Vanden Berghen's XML Parser*/
/* ------------------------------------------------------------------------- */

using std::string;
#define GM_SUPPLY_ON 0xa1
#define GM_SUPPLY_OFF 0xa0
#define RESET_COUNTERS 0xa2
#define LCD_BACKLIGHT_ON 0x11
#define LCD_BACKLIGHT_OFF 0x10
#define BUZZER_ON 0x41
#define BUZZER_OFF 0x40

static char *usbErrorMessage(int errCode)
{
    static char buffer[80];

    switch(errCode)
    {
        // by b0b3r
    case USBOPEN_ERR_ACCESS:
        return (char*)"Access to device denied";
    case USBOPEN_ERR_NOTFOUND:
        return (char*)"The specified device was not found";
    case USBOPEN_ERR_IO:
        return (char*)"Device communication error";
    default:
        sprintf(buffer, "Unknown USB error %d", errCode);
        return buffer;
    }
    return NULL;    /* not reached */
}

static usbDevice_t  *openDevice(int *error, int *revision)
{
    usbDevice_t     *dev = NULL;
    unsigned char   rawVid[2] = {USB_CFG_VENDOR_ID}, rawPid[2] = {USB_CFG_DEVICE_ID};
    char            vendorName[] = {USB_CFG_VENDOR_NAME, 0}, productName[] = {USB_CFG_DEVICE_NAME, 0};
    int             vid = rawVid[0] + 256 * rawVid[1];
    int             pid = rawPid[0] + 256 * rawPid[1];
    int             err; //"local" error
	int				rev; //sensor hardware revision
	
	err = usbhidOpenDevice(&dev, vid, vendorName, pid, productName, 0, &rev);
    if(err != 0)
    {
		if (err != *error) {
			fprintf(stderr, "error finding '%s': %s\n", productName, usbErrorMessage(err));
			*error = err; //rewriting the "local" error to the global var
		}
        return NULL;
    }
	//rewriting sensor revision to globally available var
	*revision = rev;
    return dev;
}

/* ------------------------------------------------------------------------- */

static void hexdump(char *buffer, int len)
{
    int     i;
    FILE    *fp = stdout;

    for(i = 0; i < len; i++)
    {
        if(i != 0)
        {
            if(i % 16 == 0)
            {
                fprintf(fp, "\n");
            }
            else
            {
                fprintf(fp, " ");
            }
        }
        fprintf(fp, "0x%02x", buffer[i] & 0xff);
    }
    if(i != 0)
        fprintf(fp, "\n");
}

static int  hexread(char *buffer, char *string, int buflen)
{
    char    *s;
    int     pos = 0;

    while((s = strtok(string, ", ")) != NULL && pos < buflen)
    {
        string = NULL;
        buffer[pos++] = (char)strtol(s, NULL, 0);
    }
    return pos;
}
/* ------------------------------------------------------------------------- */
unsigned int convert4bytes(char *buffer)
{
    unsigned int base = 1;
    unsigned int result = 0;
    int i;
    for (i = 1; i <= 4; i++)
    {
        result += base * ((unsigned char) buffer[i]);
        base *= 256;
    }
    return result;
}
/* ------------------------------------------------------------------------- */

static void usage(char *myName)
{
    fprintf(stderr, "usage:\n");
    fprintf(stderr, "  %s read\n", myName);
    fprintf(stderr, "  %s write <listofbytes>\n", myName);
}

struct SENSOR_HW {
	//this defines what the sensor can do.
	bool switch_backlight;
	bool switch_buzzer;
	bool switch_gmsupply;
	bool reset_counters;
	bool txt_line1; //v2 
	bool txt_line2; //obsolete now - earlier v1s only.
	unsigned char hardware_version;
	unsigned char software_version;
	char revision_txt[16];
};

SENSOR_HW get_hw_data (int rev) {
	//this function takes the integer returned by USB and returns 
	//sensor version and caps
	int hardware_version;
	int software_version;
	SENSOR_HW sensor_capabilities;
	hardware_version = (int) rev / 256;
	software_version = (int) rev - (256 * hardware_version);
	sprintf(sensor_capabilities.revision_txt,"v%d.%02d",hardware_version,software_version);
	switch (hardware_version) {
		case 2:
			sensor_capabilities.txt_line2 = false;
			sensor_capabilities.txt_line1 = true;
			sensor_capabilities.switch_gmsupply = true;
			sensor_capabilities.switch_backlight = true;
			sensor_capabilities.switch_buzzer = true;
			sensor_capabilities.reset_counters = true;
			break;
		default:
			sensor_capabilities.txt_line2 = false;
			sensor_capabilities.txt_line1 = false;
			sensor_capabilities.switch_gmsupply = false;
			sensor_capabilities.switch_backlight = false;
			sensor_capabilities.switch_buzzer = false;
			sensor_capabilities.reset_counters = true;
			strcpy(sensor_capabilities.revision_txt,"[Unknown]");
		break;
	}//switch ends here
	return sensor_capabilities;
}

int sensor_write(char command, int pos, int retry, usbDevice_t *sensor, int *usb_err) {
	//function takes command, [buffer] position and retry count as arguments
	//retry# is used to retry communication with sensor if there was an error
	//most of USB errors are not critical and it is better to repeat a couple of times
	//instead of giving up immediately.
	//pointer to usb_write_errors is used to modify *global* value of 'soft' USB write errors
	//it is used for debugging USB transfers and may be removed in future
	//
	//new local USB buffer - I made it small because we send at most 1-2 bytes anyway
	//so there's no need to transfer 127 bytes of zeros
	//(I hope that the rest of the code will handle the shorter buffer)
	char	l_buffer[129];
	memset(l_buffer, 0, sizeof(l_buffer));
	//setting command at the right position; as far as I know pos 0 is reserved for report ID
	l_buffer[pos] = command;
	//usb error value, will be used to signal errors to the main loop
	int usberr = 0;
	//the loop does <retry#> retries on failure
	for (int licznik=0;licznik<retry;licznik++) {
		usberr = usbhidSetReport(sensor, l_buffer, sizeof(l_buffer));
		if (!usberr) break;
		//increasing the number of usb errors 
		*usb_err += 1;
		boinc_sleep(1);
	}
	return usberr;	
}


int main(int argc, char **argv)
{
	/* BOINC INIT */
	APP_INIT_DATA init_data;
	boinc_init();
	/* loading data from APP_INIT_DATA */
	/* APP INIT contains the project_specific_prefs section */
	/* which is needed to read backlight/buzzer on/off status */
	boinc_get_init_data(init_data);
	/* INIT SECTION ENDS HERE */
	usbDevice_t *dev;
	//stores the # of 'soft' read/write errors. 
	int usb_write_errors = 0;
	int usb_read_errors = 0;
	//stores number of 'hard' errors
	//reset back to zero on success, used to disconnect broken connection.
	int usb_n_errors = 0;
	//debug flag, later read from prefs
	int debug = 0;
	//long workunits 
	bool long_wus = false;
	//buffer stores USB/HID I/O data. 
	char	buffer[129];    /* room for dummy report ID */
	//used to store SVN revision of this file
	//not really needed but it's printed to stderr on startup
	char	svn_revision[64]; 
	// err holds the USB transmission status
	int		err;
	//timers and counters, keeps values read from device
	unsigned int timer = 0;
	unsigned int timer_old = 0;
	unsigned int counter = 0;
	unsigned int counter_old = 0;
	//counter_sum holds the sum of all readings 
	//used to warn user about faulty GM tube
	unsigned int counter_sum = 0; 
	int		s;
	int		i=1;
	int		j=1;

	SENSOR_HW sensor; //this struct keeps all the info about device

	//repeats holds the number of main loop iterations
	//it also defines max runtime when the sensor is connected at repeats * m_interval
	//does not affect runtime if the sensor is not connected
	int	repeats=121;
	//holds the number of measures read from output file
	//(if it exists)
	//used to recalculate the progress meter
	int			n_earlier_measures = 0;
	//error holds openDevice() error status (if any). 
	//used to avoid displaying the same error message more than once
	//if the sensor is not present
	int			error = 0; 
	//this is our progress info
	double      percentage = 0.0;
	//each measurement is marked with a flag which defines it's type
	//currently there are 3 different measurements
	//"f" is the first measurement after startup
	//(it may contain data from long periods if the sensor was left on without app running,
	//or just the data since the end of previous run)
	//"n" is used to mark measurements during normal operation
	//"r" is used to mark first measurement after resume/restart
	char		flag[1]; 
	//fstart is used to inform the main loop that we're doing first reading
	bool        fstart = true;
	//this is switched to true later if the sensor is detected
	bool 		sensor_connected = false; 
	//restarted is switched to true later, after the checkpoint file
	//(if it exists) is read. 
	bool		restarted = false;
	//how long should we wait for sensor (minutes):
	int	wait_time = 300; 
	//sleep time between reading from the sensor
	//it's best to keep it divisible by 40, as it makes calculations easy 
	int	sample_time = 40; 
	//points to project_prefs which holds info needed to customize buzzer and backlight status 
	char		*project_prefs = init_data.project_preferences;
	//just to make parsing prefs easier:
	string parse_prefs = " ";
	//quick check if the prefs are actually there
	//if the user didn't change the prefs, project_prefs is empty
	if (project_prefs) {
		parse_prefs = project_prefs;
	}
	//sensor's software/hardware revision
	//the app needs to know what the sensor can do
	//to avoid sending commands which the sensor can't understand 
	//newer revisions are (usually) backwards compatible
	int sensor_rev = 0;

	//stores the trickle msg buffer 
	string trickle_msg_buff;
	//stores the single trickle msg (it is later copied to buff)
	char trickle_msg[1024];
	//the number of trickles in buffer - used to decide
	//if it's time to contact the server
	int n_trickles = 0;
	//defines the # of trickles in buffer before contacting the server
	//affects directly the connection frequency
	const int max_trickles = 30;
	//standalone/boinc mode switch
	//needed because older BOINC API had problems
	//with trickles when running standalone
	bool standalone = boinc_is_standalone();

	strcpy(flag,"f"); 
	strcpy(svn_revision,LAST_CHANGED_REVISION);

    time_t currTime;
    struct tm* t;

	/* APP PREFS */
	//this block deals with things related to application preferences
	//poor man's XML parser
	/* long workunits */
	if (parse_prefs.find("<longwu>1</longwu>") != parse_prefs.npos) {
		//longwu increases the WU runtime
		repeats *= 4;
		long_wus = true;
		//note: if user changes his prefs while the app is off,
		//it may restart with the number of measurements higher than maximum
		//it is handled by the init block (the app will exit)
	}
	/* debug */
	if (parse_prefs.find("<radacdebug>1</radacdebug>") != parse_prefs.npos) {
		//if the debug is set in prefs, override the app timings and set the debug flag
		//it also overrides some other prefs settings
		debug = 1;
		sample_time = 20;
		repeats = 8;
		//Prefs/Long workunits also affects debug WUs
		if (long_wus) repeats = 80;
		wait_time = 30;
		fprintf(stderr, "Debug: enabled; sample time: %d, repeats: %d \n", sample_time,repeats);
	}

	fprintf(stderr,"Radac %s starting...\n",svn_revision);
	//here the app looks for output file and reads it if it exists
	FILE *pFile;
	pFile=boinc_fopen("data.bin","r");
	if (pFile) {
		//dirty way to read number of previous measurements from file
		//however it works if the file structure is ok
		//if it's not - doesn't really matter
		//validator will take care of this anyway
		char str[2048];
		char *pch;
		while (!feof(pFile)) { 
			fgets(str,2048,pFile);
			n_earlier_measures++;
		}
		fclose(pFile);
		//if the number of measures read from output file is close to the max,
		//it's better to finish the WU here and start a new one
		if (n_earlier_measures > repeats - 6) {
			//calling API just to signal the core that app did something
			boinc_checkpoint_completed();
			boinc_fraction_done(0.99); 
			fputs("Found previously created output file, workunit almost completed - exiting\n",stderr);
			boinc_finish(0);
		}
		//if the # of measures was too low, app resumes here.
		//'restarted' overrides 'fstart' 
		restarted = true;
		fstart = false;
		char dane[7][256];
		//"str" keeps the last line read from checkpoint file, it'll be split here 
		//because the timer from last reading is needed later
		//to determine if the app was paused for a short time, or restarted
		pch = strtok (str,",\n");
		int n = 0;
		while (pch != NULL) {
			n++;
			strcpy(dane[n],pch);
			pch = strtok (NULL, ",\n");
		}
		//n=6 means that the line structure was probably ok. 
		//timer_old is set to the last value read from file
		if (n == 6) timer_old = atoi(dane[1]); 
	}

/*	At this point the app is fully initialized
	so it's time to start looking for sensor */

	for (int tmp_counter=wait_time;tmp_counter>0;tmp_counter--) {
		//"error" is passed to openDevice; it contains the last error that occured
		//if it fails the same way again (for example: no hardware sensor)
		//openDevice() won't write to stderr. 
		if((dev = openDevice(&error, &sensor_rev)) == NULL) {
			//on error sleep 60 seconds before retry
			sensor_connected = false;
			for (int z=0;z<60;z++) {
				boinc_sleep(1); 
			}
		}
		else {
			sensor_connected = true;
			//get the device version and capabilities from the version string.
			sensor = get_hw_data(sensor_rev);
			fprintf(stderr,"Found sensor %s\n",sensor.revision_txt);
			//sensor found, app leaves this loop here
			break; 
		}
	}
    if (!sensor_connected) { 
		//no communication with the sensor, the app will now exit
		FILE *pFile;
		pFile = boinc_fopen("data.bin","ab");
		if (!pFile) {
			fputs("Error opening pFile\n",stderr);	
			exit(EXIT_FAILURE);
		}
		//write to output file, the validator pick this up 
		//and marks the result as valid with 0 credit
		s=fprintf(pFile,"No detector connected");
		fclose (pFile);
		boinc_finish(0);
	};



	int len = sizeof(buffer);
	//the buffer is cleared before first use
	memset(buffer, 0, len);

	/* H/W DEBUG */
	if (debug) {
		//If debug flag is set, we go through all the sensor commands 
		//to see if everything is working/responding
		//counters are reset
		//because switching off the GM supply affects the readings
		if (sensor.switch_gmsupply) err = sensor_write(GM_SUPPLY_OFF,1,10,dev,&usb_write_errors);
		boinc_sleep(1);
		if (sensor.switch_buzzer) err = sensor_write(BUZZER_ON,1,10,dev,&usb_write_errors);
		boinc_sleep(1);
		if (sensor.switch_buzzer) err = sensor_write(BUZZER_OFF,1,10,dev,&usb_write_errors);
		boinc_sleep(1);
		if (sensor.switch_gmsupply) err = sensor_write(GM_SUPPLY_ON,1,10,dev,&usb_write_errors);
		boinc_sleep(1);
		if (sensor.reset_counters) err = sensor_write(RESET_COUNTERS,1,10,dev,&usb_write_errors);
		boinc_sleep(1);
		if (sensor.switch_backlight) err = sensor_write(LCD_BACKLIGHT_OFF,1,10,dev,&usb_write_errors);
		boinc_sleep(1);
		if (sensor.switch_backlight) err = sensor_write(LCD_BACKLIGHT_ON,1,10,dev,&usb_write_errors);
		boinc_sleep(1);
		if (sensor.switch_backlight) err = sensor_write(LCD_BACKLIGHT_OFF,1,10,dev,&usb_write_errors);
		boinc_sleep(1);
		if (sensor.switch_backlight) err = sensor_write(LCD_BACKLIGHT_ON,1,10,dev,&usb_write_errors);
		boinc_sleep(1);
		if (sensor.switch_backlight) err = sensor_write(LCD_BACKLIGHT_OFF,1,10,dev,&usb_write_errors);
		boinc_sleep(1);
		if (sensor.switch_backlight) err = sensor_write(LCD_BACKLIGHT_ON,1,10,dev,&usb_write_errors);
		boinc_sleep(1);
		if (sensor.switch_backlight) err = sensor_write(LCD_BACKLIGHT_OFF,1,10,dev,&usb_write_errors);
		boinc_sleep(1);
	}
	/* SENSOR PREFS */
	//this block deals with things related to sensor preferences
	//since revision 241 it respects the sensor hardware capabilities
	//poor man's XML parser
	/* buzzer */
	if (sensor.switch_buzzer) { //do not try to switch if the sensor can't do this
		if (parse_prefs.find("<buzzer>0</buzzer>") != parse_prefs.npos) {
			err = sensor_write(BUZZER_OFF,1,10,dev,&usb_write_errors);
		}
		else {
			err = sensor_write(BUZZER_ON,1,10,dev,&usb_write_errors);
		}
		//sensor_write() returns USB error *only* if all retries failed.
		if (err) {
			usb_n_errors++;
			fprintf(stderr, "Error writing data [buzzer on/off]: %s\n", usbErrorMessage(err));
		}
	}
	/* backlight */
	// and once more for backlight status
	if (sensor.switch_backlight) { //do not try to switch if the sensor can't do this
		if (parse_prefs.find("<backlight>0</backlight>") != parse_prefs.npos) {
			err = sensor_write(LCD_BACKLIGHT_OFF,1,10,dev,&usb_write_errors);
		}
		else {
			err = sensor_write(LCD_BACKLIGHT_ON,1,10,dev,&usb_write_errors);
		}
		//sensor_write() returns USB error *only* if all retries failed.
		if (err) {
			usb_n_errors++;
			fprintf(stderr, "Error writing data [backlight on/off]: %s\n", usbErrorMessage(err));
		}
	}
	/* MAIN LOOP */
	//the main loop is initialized with number of measures done earlier
	//it limits the number of iterations left and also sets the progress bar to correct value
	for (i = n_earlier_measures; i < repeats; i++) {
		//calculating progress bar
		percentage = (double) i / repeats;
		//It is safer to avoid stopping the app while it's writing to checkpoint
		//and/or reading from USB, so it enters boinc critical section here.
		boinc_begin_critical_section(); 
		//from time to time USB communication ends with error
		//a single error is not critical, however it is safe to retry a couple of times
		//and print error message only if all retries failed
		for (int licznik=0;licznik<10;licznik++) {
			err = usbhidGetReport(dev, 0, buffer, &len);
			//err=0 means we got reading from USB, exiting loop
			if (!err) {
				//resetting number of hard errors back to 0
				usb_n_errors = 0;
				break;
			}
			usb_read_errors++;
			boinc_sleep(1);
		}
		//on error do not enter the processing block
		//missing one reading doesn't affect anything
		//however missing a couple of readings in a row might indicate 
		//communication problems
		if (err) {
			usb_n_errors++;
			fprintf(stderr, "Error reading data: %s\n", usbErrorMessage(err));
			//If there were more than 3 errors in a row, close the device and exit
			if (usb_n_errors > 3) {
				boinc_end_critical_section();
				boinc_fraction_done(0.99);
				usbhidCloseDevice(dev);
				//do we have any trickles not sent yet?
				if (n_trickles > 0 && !standalone) {
					err = boinc_send_trickle_up("rad_report", const_cast< char* > (trickle_msg_buff.c_str()));
					if (!err) {
						//(don't care about buffer, app is exiting anyway
						if (debug) fprintf(stderr, "Debug: trickle sent\n");
					}
					else {
						if (debug) fprintf(stderr, "Debug: boinc_send_trickle_up() returned an error\n");
					}
				}
				fprintf(stderr, "Device not responding, exiting.... \n");
				boinc_finish(0);
			}
		}
		else {
			/* Data processing block */ 
			timer = convert4bytes(buffer);
			counter = convert4bytes(buffer + 4);
			counter_sum += counter;
			//trying to open checkpoint/data file
			//exit [with error msg] on error
			FILE *pFile;
			if ((pFile=boinc_fopen("data.bin","ab"))==NULL) {
				fputs("Error opening pFile\n",stderr);
				exit(EXIT_FAILURE);
			}
			//the timer and counter are compared with their previous values
			//easy way to detect various malfunctions
			if (timer != timer_old || counter != counter_old) {
				//dealing with flags 
				if (fstart == true) {
					strcpy(flag,"f");
					fstart = false;
				}
				else {
					if (restarted == true) {
						strcpy(flag,"r");
						//the app was restarted or unpaused; [72000000 = ~120 minutes]
						//if the app was off for less than 120 minutes (but the sensor was working)
						//mark the reading as "normal" otherwise it's safer to leave "resumed" flag
						if (timer > timer_old && timer - timer_old < 72000000) {
							strcpy(flag,"n");
							if (debug) fprintf(stderr, "Debug: app restarted/resumed, sensor was running");
						}
						restarted = false;
					}
					else {
						strcpy(flag,"n");
					}
				}
				currTime = time(NULL);
				t = gmtime(&currTime);
				//to be cleaned later
				fprintf(pFile,"%u,%u,%u-%u-%u %u:%u:%u,%i,%s,%f\n",(unsigned int)timer,(unsigned int)counter,t->tm_year+1900,t->tm_mon+1,t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec, j, flag, percentage);
				fprintf(stderr,"%u,%u,%u-%u-%u %u:%u:%u,%i,%s,%f\n",(unsigned int)timer,(unsigned int)counter,t->tm_year+1900,t->tm_mon+1,t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec, j, flag, percentage);
				if (!standalone) { 
					//trickle handling crashes app if running standalone
					//or at least it was before recent fix in the API
					sprintf(trickle_msg,"%u,%u,%u-%u-%u %u:%u:%u,%i,%s,%f\n",(unsigned int)timer,(unsigned int)counter,t->tm_year+1900,t->tm_mon+1,t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec, j, flag, percentage);
					//adding the trickle to trickle buff and increasing the trickle count
					trickle_msg_buff.append(trickle_msg);
					n_trickles++;
					//do we have enough trickles to send data up ?
					if (n_trickles > max_trickles) {
						//The const_cast may be a little risky, however the buffer is zeroed later anyway
						err = boinc_send_trickle_up("rad_report", const_cast< char* > (trickle_msg_buff.c_str()));
						if (!err) {
							//clear buffer at success and set the # of trickles to 0
							trickle_msg_buff = "";
							n_trickles = 0;
							if (debug) fprintf(stderr, "Debug: trickle sent\n");
							//append the last message again but do not increase n_trickles
							//the last entry is needed to get initial value for calculations
							//because trickle_msg may start somewhere in the middle (without "f" flag)
							//and then the trickle handler doesn't have previous data
							trickle_msg_buff.append(trickle_msg);
						}
						else {
							if (debug) fprintf(stderr, "Debug: boinc_send_trickle_up() returned an error\n");
						}
					}
					//clear the trickle msg (it's already in buff)
					strcpy (trickle_msg," ");
				}
				j++;
				fstart = false;
			}
			//this block is done if we got a reading
			//even if the reading was invalid (faulty tube, etc)
			fclose (pFile);
			counter_old = counter;
			timer_old = timer;
			boinc_checkpoint_completed();
		}
		//this is always done:
		//boinc critical section was open before trying to get data from USB
		//so it must be closed here (even if the USB transfer failed)
		boinc_end_critical_section();
		boinc_fraction_done(percentage);
		//the wait time between two measurements
		for (int licznik=0;licznik<sample_time;licznik++) {
			boinc_sleep(1);
		}
	}
	//^^^^^this is the end of the main loop
	//do we have any trickles not reported yet?
	if (n_trickles > 0 && !standalone) {
		err = boinc_send_trickle_up("rad_report", const_cast< char* > (trickle_msg_buff.c_str()));
		if (!err) {
			//clear buffer at success and set the # of trickles to 0
			//doesn't matter here anyway because we're exiting 
			trickle_msg_buff = "";
			n_trickles = 0;
			if (debug) fprintf(stderr, "Debug: trickle sent\n");
		}
		else {
			if (debug) fprintf(stderr, "Debug: boinc_send_trickle_up() returned an error\n");
		}
	}
	//warn about possibly broken GM tube
	if (counter_sum == 0) fputs("WARNING: No readings from GM tube !\n",stderr);
	//the sensor counters are reset to 0 at exit
	if (sensor.reset_counters) err = sensor_write(RESET_COUNTERS,1,10,dev,&usb_write_errors);
	if (err) fprintf(stderr, "Error writing data: %s\n", usbErrorMessage(err));
	if (debug) fprintf(stderr, "Debug: number of USB errors read: %d write: %d\n", usb_read_errors, usb_write_errors);
	fputs("Done - calling boinc_finish()\n",stderr);
	usbhidCloseDevice(dev);
	boinc_finish(0);
    return 0;
}

/* ------------------------------------------------------------------------- */
